#
# yapsnmp - Yet Another Python SNMP module
#
# This is a highish level Python SNMP module, based on my SWIGed net-snmp
# (formerly known as ucd-snmp) interface. It currently supports v1 and v2c,
# with full MIB parsing. v3 coming soon (shouldn't be technically a problem,
# simply a matter of finding the time to implement it).
#
#  -- Yves Perrenoud <yves@njc.ch>

__version__ = "0.7"

import netsnmp, socket, string

# initialize the library (mostly involves parsing in MIBs)

netsnmp.init_snmp("snmpapp")

# define supported versions of SNMP

snmpversions = {1: netsnmp.SNMP_VERSION_1,
                2: netsnmp.SNMP_VERSION_2c,
                }

# define module specific exceptions

class MIBParseError(Exception): pass
class GetError(Exception): pass
class VersionMismatch(Exception): pass
class OIDTypeError(Exception): pass

# a sequencing testing function

def issequence(obj):
    if type(obj) == type(()) or type(obj) == type([]):
        return 1
    else:
        return 0

# function returning a clean string without surrounding double quotes

def cleanval(str):
    if str and (str[0] == str[-1] == '"'):
        return str[1:-1]
    else:
        return str

# function transforming an ASCII OID to the appropriate oid object

def get_node(reqoid):
    if reqoid[:2] == '.1':
        reqoid = 'iso' + reqoid[2:]

    oid = netsnmp.py_get_node(reqoid)
    if oid.this == 'NULL':
        raise MIBParseError, reqoid

    return oid

# function translating a numeric or partial OID, to a fully qualified version

def translate(reqoid):
    return netsnmp.py_sprint_objid(get_node(reqoid))

# function toggling numeric representations of OIDs

def toggle_numeric_oids():
    netsnmp.ds_toggle_boolean(netsnmp.DS_LIBRARY_ID, netsnmp.DS_LIB_PRINT_NUMERIC_OIDS)

# The main Class allowing the SNMP interaction

class Session:
    def __init__(self, peername, version=1, community='public'):
        session = netsnmp.snmp_session()
        netsnmp.snmp_sess_init(session)

        if not peername:
            raise ValueError, "expecting a non null string as host identifier"

        if socket.gethostbyname(peername):
            session.peername = peername

        self.version = version
        session.version = snmpversions[version]

        if type(community) == type(''):
            session.community = community
            session.community_len = len(community)
        else:
            raise TypeError, "expecting a community of type string"

        self.session = netsnmp.snmp_sess_open(session)

    def __del__(self):
        try:
            netsnmp.snmp_sess_close(self.session)
        except AttributeError:
            pass

    def __sprint__(self, variable, printoid):
        if printoid:
            value = string.split(netsnmp.py_sprint_variable(variable), ' = ')
            value[1] = cleanval(value[1])
            value = tuple(value)
        else:
            value = cleanval(netsnmp.py_sprint_value(variable))

        return value

    def __synch_response__(self, pdu, printoid=0):
        response = netsnmp.py_snmp_sess_synch_response(self.session, pdu)
        if response.this != 'NULL' and response.errstat == netsnmp.SNMP_ERR_NOERROR:
            if (not self.__sequence__) and (response.variables.next_variable.this == 'NULL'):
                value = self.__sprint__(response.variables, printoid)
            else:
                value = ()
                curvar = response.variables
                while curvar.this != 'NULL':
                    value = value + (self.__sprint__(curvar, printoid),)
                    curvar = curvar.next_variable

            netsnmp.snmp_free_pdu(response)
            return value
        else:
            if response.this == 'NULL':
                errmsg = "couldn't obtain response"
            else:
                errmsg = netsnmp.snmp_errstring(response.errstat)
                netsnmp.snmp_free_pdu(response)

            raise GetError, errmsg

    def __fill_get_pdu__(self, pdu, oidlist):
        if issequence(oidlist[0]):
            oidlist = oidlist[0]
            self.__sequence__ = 1
        else:
            self.__sequence__ = 0

        for argoid in oidlist:
            oid = get_node(argoid)
            netsnmp.snmp_add_null_var(pdu, oid.oid, oid.len)

    def get(self, reqoid, *args):
        pdu = netsnmp.snmp_pdu_create(netsnmp.SNMP_MSG_GET)
        self.__fill_get_pdu__(pdu, (reqoid,) + args)
        return self.__synch_response__(pdu)

    def getnext(self, reqoid, *args):
        pdu = netsnmp.snmp_pdu_create(netsnmp.SNMP_MSG_GETNEXT)
        self.__fill_get_pdu__(pdu, (reqoid,) + args)
        return self.__synch_response__(pdu, printoid=1)

    def getbulk(self, reqoid, *args, **kw):
        if self.version == 1:
            raise VersionMismatch, "GETBULK requires v2c and higher"

        pdu = netsnmp.snmp_pdu_create(netsnmp.SNMP_MSG_GETBULK)

        try:
            pdu.errstat = kw['non_repeaters']
        except KeyError:
            pdu.errstat = 0

        try:
            pdu.errindex = kw['max_repetitions']
        except KeyError:
            pdu.errindex = 100

        self.__fill_get_pdu__(pdu, (reqoid,) + args)
        return self.__synch_response__(pdu, printoid=1)

    def set(self, reqoid, *args):
        if not issequence(reqoid):
            raise TypeError, "expecting a sequence"

        pdu = netsnmp.snmp_pdu_create(netsnmp.SNMP_MSG_SET)

        if issequence(reqoid[0]):
            oidlist = reqoid
            self.__sequence__ = 1
        else:
            oidlist = (reqoid,) + args
            self.__sequence__ = 0

        for argoid in oidlist:
            oid = get_node(argoid[0])
            netsnmp.snmp_add_var(pdu, oid.oid, oid.len, argoid[1], argoid[2])

        return self.__synch_response__(pdu, printoid=1)

    def walk(self, reqoid):
        lreqoid = translate(reqoid)
        walklist = []
        nextoid = lreqoid

        while nextoid:
            if self.version == 1:
                oidtuplelist = [self.getnext(nextoid)]
            else:
                oidtuplelist = self.getbulk(nextoid)

            for oidtuple in oidtuplelist:
                nextoid = oidtuple[0]
                if lreqoid == nextoid[:len(lreqoid)]:
                    oidtuple = (reqoid + oidtuple[0][len(lreqoid):], oidtuple[1])
                    walklist.append(oidtuple)
                else:
                    nextoid = ''
                    break

        return tuple(walklist)

    def table(self, reqoid):
        walktuple = self.walk(reqoid)

        if not walktuple:
            raise OIDTypeError, "expecting a table"

        posoid = string.split(walktuple[0][0], '.').index(reqoid)

        headers = []
        tabledict = {}
        table = []

        for oidtuple in walktuple:
            oidlist = string.split(oidtuple[0], '.')

            if oidlist[posoid + 1][-5:] != "Entry":
                raise OIDTypeError, "expecting a table"

            index = string.join(oidlist[posoid + 3:], '.')
            varname = oidlist[posoid + 2]

            if varname not in headers:
                headers.append(varname)

            if not tabledict.has_key(index):
                tabledict[index] = {}

            tabledict[index][varname] = oidtuple[1]

        indexes = tabledict.keys()
        indexes.sort()
        for index in indexes:
            table.append(tabledict[index])

        return tuple(table)
